Option Compare Binary
Option Explicit On 
Option Strict On

Public Enum ECaste
    cstWorker
    cstDrone
    cstQueen
End Enum

' The different kinds of motion that bees can have.
Public Enum EMotion
    ' Travelling back to the hive (only relevant if the bee is outside).
    motHoming
    ' Randomly searching for food outside.
    motSearching
    ' Returning to a known food source.
    motDefinite
    ' Not moving (probably a new-born bee).
    motStill
End Enum

Public MustInherit Class CBee

    ' The ID of the hive that this bee belongs to.
    Protected mintHiveID As Integer = 0

    ' The age (in ticks).
    Protected mintAge As Integer = 0

    ' Is this bee inside or outside the hive?
    ' All bees are born inside the hive, and may leave it later.
    Protected mblnOutside As Boolean = False

    ' The comb index, if the bee is inside the hive.
    Protected mintComb As Integer = 0

    ' The position of the bee on the relevant plane.
    ' This plane is either the ground (if it's outside), or a vertical comb surface (if it's inside).
    ' TODO: Track the horizontal position of the bee if it is flying underneath all the combs to get to the exit of the hive.
    ' For now, just let the bee "teleport" between a comb and the exit, using the magic of quantum tunnelling.
    Protected mptBee As New PointF(0, 0)

    ' The current direction that the bee is travelling in.
    ' Measured in radians, so -Pi < x <= Pi.
    ' 0 = right.
    Protected mdblDirection As Double = 0

    ' The speed that bees travel.
    ' For now, assume it's the same for all bees.
    ' TODO: Get accurate figures (with biblio reference!) for this.
    Protected Const mcintCrawlingSpeed As Integer = 20  ' mm/sec
    ' Cruising speed = 25km/hour = 25,000,000mm/3600 seconds = 6944mm/second (circa 7m/sec).
    Protected Const mcintFlyingSpeed As Integer = 7

    ' The speed that bees travel, in mm/m per tick.
    ' For now, assume it's the same for all bees.
    Protected mintCrawlingSpeed As Integer = CType(mcintCrawlingSpeed * CDome.SpeedMultiplier, Integer)
    Protected mintFlyingSpeed As Integer = CType(mcintFlyingSpeed * CDome.SpeedMultiplier, Integer)

    ' What kind of motion does this bee have?
    Protected mmotThis As EMotion = EMotion.motStill

    ' If this bee is aiming for something in particular (food source/hive),
    ' store the target's position.
    Protected mptTarget As PointF

    ' The location of this bee's "home" hive.
    Protected mptHive As PointF

    ' The distance to move in each tick.
    Private msngOffsetX As Single = 0
    Private msngOffsetY As Single = 0

    ' How long can this bee live without eating?
    ' The (constant) threshold is measured in days,
    ' and the variable is measured in ticks.
    Protected Const mcintStarvationThreshold As Integer = 1
    Protected mintNextMealDue As Integer

    ' What is the minimum (dome) temperature that a bee will fly in? Measured in Kelvins.
    Protected Const mcintMinTemperature As Integer = 278

    ' The bee has died.
    Public Event DeadBee(ByVal beeThis As CBee)

    ' The bee has left a hive (to enter the dome area).
    Public Event LeftHive(ByVal beeThis As CBee)

    ' The bee has left the dome area (to enter a hive).
    Public Event LeftDome(ByVal beeThis As CBee)

    Public Property HiveID() As Integer
        Get
            Return mintHiveID
        End Get
        Set(ByVal Value As Integer)
            mintHiveID = Value
        End Set
    End Property

    Public ReadOnly Property Location() As PointF
        Get
            Return mptBee
        End Get
    End Property

    Public Overridable Sub Tick()

        ' Adjust the age.
        mintAge += 1

        ' Only adjust the hunger if the bee is outside.
        ' (Assume that any bees in a hive have permanent access to food.)
        If mblnOutside Then
            mintNextMealDue -= 1
        End If

        ' Kill/move this bee if appropriate.
        ' N.B. "Still" includes hovering in position, e.g. waiting to mate.
        ' Sub-classes can add specific behaviour when they override Tick.
        If (mintAge > LifeExpectancy * CDome.TicksPerDay) Or (mintNextMealDue < 0) Then
            Die()
        ElseIf (mmotThis <> EMotion.motStill) Then
            Move()
        End If

    End Sub

    ' The caste.
    Public MustOverride ReadOnly Property Caste() As ECaste

    ' The bee's life expectancy (in days).
    Public MustOverride ReadOnly Property LifeExpectancy() As Integer

    Public ReadOnly Property Motion() As EMotion
        Get
            Return mmotThis
        End Get
    End Property

    Public ReadOnly Property Direction() As Double
        Get
            Return mdblDirection
        End Get
    End Property

    Protected Sub CalculateOffset()
        ' This stores the speed.
        Dim intSpeed As Integer

        ' Get the relevant speed.
        If mblnOutside Then
            intSpeed = mintFlyingSpeed
        Else
            intSpeed = mintCrawlingSpeed
        End If

        ' Calculate the offset.
        ' N.B. We need special cases for the 4 "basic" directions (up, down, left, right),
        ' since one of the components (either x or y) will be zero in each situation,
        ' so we can't use trigonometry to calculate the lengths of sides of a right-angled triangle.
        Select Case mdblDirection
            Case -Math.PI / 2
                ' Down.
                msngOffsetX = 0
                msngOffsetY = -intSpeed
            Case 0
                ' Right.
                msngOffsetX = intSpeed
                msngOffsetY = 0
            Case Math.PI / 2
                ' Up.
                msngOffsetX = 0
                msngOffsetY = intSpeed
            Case Math.PI
                ' Left.
                msngOffsetX = -intSpeed
                msngOffsetY = 0
            Case 0 To Math.PI / 2
                ' Up-right.
                msngOffsetX = CInt(Math.Cos(mdblDirection) * intSpeed)
                msngOffsetY = CInt(Math.Sin(mdblDirection) * intSpeed)
            Case Math.PI / 2 To Math.PI
                ' Up-left.
                msngOffsetX = -CInt(Math.Cos(Math.PI - mdblDirection) * intSpeed)
                msngOffsetY = CInt(Math.Sin(Math.PI - mdblDirection) * intSpeed)
            Case -Math.PI / 2 To 0
                ' Down-right.
                msngOffsetX = CInt(Math.Cos(mdblDirection) * intSpeed)
                msngOffsetY = CInt(Math.Sin(mdblDirection) * intSpeed)
            Case -Math.PI To -Math.PI / 2
                ' Down-left.
                msngOffsetX = -CInt(Math.Cos(Math.PI - mdblDirection) * intSpeed)
                msngOffsetY = CInt(Math.Sin(Math.PI - mdblDirection) * intSpeed)
            Case Else
                msngOffsetX = 0
                msngOffsetY = 0
        End Select

    End Sub

    ' Fly towards the target.
    Protected Sub Aim()
        ' Temporary angle for calculation.
        Dim dblPhi As Double

        ' Pick the direction that leads back towards the target.

        ' Get the angle between the x-axis and a vector from the target to this bee's position.
        ' This is the direction a bee would use to travel from the target to this point.
        ' N.B. This angle is in radians.
        dblPhi = Math.Atan2(mptBee.Y - mptTarget.Y, mptBee.X - mptTarget.X)

        ' Now get the opposite direction.
        If (dblPhi > 0) Then
            ' Above the x-axis.
            mdblDirection = -(Math.PI - dblPhi)
        Else
            ' Below the x-axis.
            mdblDirection = Math.PI + dblPhi
        End If

    End Sub

    Protected Sub Move()

        ' Pick a direction to move in.
        Select Case mmotThis
            Case EMotion.motDefinite, EMotion.motHoming
                ' Aim for the target.
                ' N.B. We need to do this in each tick, not just at the start,
                ' otherwise we get cumulative error problems.
                ' Besides, this is probably a more accurate model of the way a real bee would do it.
                Aim()
            Case EMotion.motSearching
                If mblnOutside Then
                    ' Either keep going in the current direction, or pick a random direction.
                    ' The figure in the If statement represents the probability of changing direction.
                    If (Rnd() < 0.5) Then
                        mdblDirection = (Rnd() * 2 * Math.PI) - Math.PI
                    End If
                Else
                    If mptBee.X < 10 Then
                        mdblDirection = 0
                    ElseIf mptBee.X > 100 Then
                        mdblDirection = Math.PI
                    ElseIf mptBee.Y < 10 Then
                        mdblDirection = -Math.PI / 2
                    ElseIf mptBee.Y > 100 Then
                        mdblDirection = Math.PI / 2
                    End If
                End If
        End Select

        ' Calculate the offset for this direction.
        CalculateOffset()

        ' Use the offset to assign the new position.
        mptBee.X += msngOffsetX
        mptBee.Y += msngOffsetY

        ' Head towards food/hive as appropriate, if outside the hive.
        If mblnOutside Then
            Select Case mmotThis
                Case EMotion.motDefinite, EMotion.motHoming
                    ' If this bee is aiming in a definite direction, see whether it has reached its target.
                    ' N.B. Allow for the possibility that it only takes "half a tick" to get there,
                    ' i.e. that it would overshoot by going the whole distance.
                    If ((Math.Abs(mptBee.X - mptTarget.X) < msngOffsetX) Or (Math.Abs(mptBee.Y - mptTarget.Y) < msngOffsetY)) Then
                        If (mmotThis = EMotion.motDefinite) Then
                            ReachTarget()
                        Else
                            EnterHive(CDome.GetHive(mintHiveID))
                        End If
                    End If
                Case EMotion.motSearching
                    Search()
            End Select
            ' We have an If statement distinct from the Case statement above,
            ' since a bee that was searching before may now be aiming for a specific target.
            If Not (mmotThis = EMotion.motSearching) Then
                Aim()
            End If
        End If  ' end "outside?" if statement.

    End Sub

    Protected Overridable Sub LeaveHive()

        ' Set the "outside?" flag.
        mblnOutside = True

        ' In dome co-ordinates, this bee is positioned at the hive.
        mptBee = mptHive

        ' Raise an event to say that this bee has left the hive.
        ' (This will cause the hive to adjust its quantities.)
        RaiseEvent LeftHive(Me)

        ' Add this bee to the dome's jurisdiction.
        CDome.AddBee(Me)

    End Sub

    Protected Overridable Sub EnterHive(ByVal hvThis As CHive)

        ' Drop the "outside?" flag.
        mblnOutside = False

        ' Leave the dome's jurisdiction.
        RaiseEvent LeftDome(Me)

        ' Come under the hive's jurisdiction.
        ' TODO: Pick a comb at random.
        hvThis.Combs(0).AddBee(Me)

        ' Start out on the corner of this comb.
        mptBee = New PointF(0, 0)

        ' It is now still (not homing).
        mmotThis = EMotion.motStill

        ' The bee can "refuel" at this point.
        Eat()

    End Sub

    Public Sub New(ByVal ptHive As PointF, ByVal ptBee As PointF, ByVal intHiveID As Integer, Optional ByVal blnOutside As Boolean = False)
        MyBase.New()
        mptHive = ptHive
        mptBee = ptBee
        mintHiveID = intHiveID
        mblnOutside = blnOutside
        If mblnOutside Then
            mmotThis = EMotion.motSearching
        Else
            mmotThis = EMotion.motStill
        End If
        Eat()
    End Sub

    Protected Sub Die()

        ' Raise the event.
        ' N.B. This seems like a long way around, but it is essential.
        ' A sub-class (e.g. CDrone) can't raise the event directly,
        ' but it can call this method.
        RaiseEvent DeadBee(Me)

    End Sub

    Protected Sub Eat()

        ' Reset the "next meal due" value to the starvation threshold.
        mintNextMealDue = mcintStarvationThreshold * CDome.TicksPerDay

    End Sub

    Protected MustOverride Sub Search()

    Protected MustOverride Sub ReachTarget()

End Class
