Public Class CDisplay

#Const cccBoundingCircle = False

    ' These constants store sizes (in mm).

    ' A standard 10 frame Langstroth hive is approximately 56cm long x 46cm wide.
    Private Const mcintHiveWidth As Integer = 460
    Private Const mcintHiveLength As Integer = 560

    ' A feeding station can be any size, so I've arbitrarily chosen a 10cm x 10cm square.
    Private Const mcintFeedingStationEdge As Integer = 100

    ' Obviously there will be some variation between individual bees, but these are averages for castes.
    Private Const mcintDroneLength As Integer = 17
    Private Const mcintQueenLength As Integer = 20
    Private Const mcintWorkerLength As Integer = 15

#Region "Dome view"
    Private Shared Sub AdjustCTM(ByVal g As Graphics, ByVal sngScaleFactor As Single, ByVal ptOrigin As PointF, ByVal ptThis As PointF)

        ' Adjust the co-ordinate transformation matrix.
        g.ResetTransform()
        g.TranslateTransform(ptOrigin.X, ptOrigin.Y)
        g.ScaleTransform(sngScaleFactor, -sngScaleFactor)
        g.TranslateTransform(ptThis.X, ptThis.Y)

    End Sub

    Private Shared Sub DrawHive(ByVal g As Graphics, ByVal sngScaleFactor As Single, ByVal sngMagnification As Single, ByVal ptOrigin As PointF, ByVal ptThis As PointF)
        ' This accesses the brush we use to draw the hive.
        Dim brHive As Brush
        ' These store the dimensions of the hive.
        Dim intWidth As Integer
        Dim intLength As Integer
        Dim rcHive As Rectangle

        ' Get the dimensions of the hive.
        ' N.B. We have to convert from millimetres to metres.
        intWidth = CType(Math.Ceiling((mcintHiveWidth / 1000) * (sngMagnification / sngScaleFactor)), Integer)
        intLength = CType(Math.Ceiling((mcintHiveLength / 1000) * (sngMagnification / sngScaleFactor)), Integer)
        rcHive = New Rectangle(0, 0, intWidth, intLength)

        ' Adjust the starting position, so that the hive's location will be based on the
        ' front door (middle of the base of the rectangle) rather than the bottom-left corner.
        ' So, we want the rectangle's bottom-left corner to be further over to the left.
        ptThis.X -= CType(intWidth / 2, Single)

        ' Adjust the co-ordinate transformation matrix.
        AdjustCTM(g, sngScaleFactor, ptOrigin, ptThis)

        ' Draw the rectangle.
        brHive = New SolidBrush(Color.White)
        g.FillRectangle(brHive, rcHive)
        brHive.Dispose()

    End Sub

    Private Shared Sub DrawBee(ByVal g As Graphics, ByVal sngScaleFactor As Single, ByVal sngMagnification As Single, ByVal ptOrigin As PointF, ByVal ptThis As PointF, ByVal motMotion As EMotion, ByVal cstCaste As ECaste)
        ' This accesses the brush we use to draw the bee.
        Dim brBee As Brush
        Dim rcBee As Rectangle
        ' This stores the diameter of a bee's bounding circle.
        Dim intDiameter As Integer

        ' Adjust the co-ordinate transformation matrix.
        AdjustCTM(g, sngScaleFactor, ptOrigin, ptThis)

        ' Create a new brush in the relevant colour for this caste.
        Select Case motMotion
            Case EMotion.motDefinite
                brBee = New SolidBrush(Color.Red)
            Case EMotion.motHoming
                brBee = New SolidBrush(Color.Blue)
            Case EMotion.motSearching
                Select Case cstCaste
                    Case ECaste.cstDrone
                        brBee = New SolidBrush(Color.Wheat)
                    Case ECaste.cstQueen
                        brBee = New SolidBrush(Color.Yellow)
                    Case ECaste.cstWorker
                        brBee = New SolidBrush(Color.LightYellow)
                End Select
        End Select

        ' Get the size of the bee.
        ' N.B. We have to convert from millimetres to metres.
        Select Case cstCaste
            Case ECaste.cstDrone
                intDiameter = mcintDroneLength
            Case ECaste.cstQueen
                intDiameter = mcintQueenLength
            Case ECaste.cstWorker
                intDiameter = mcintWorkerLength
        End Select
        intDiameter = CType(Math.Ceiling((intDiameter / 1000) * (sngMagnification / sngScaleFactor)), Integer)
        rcBee = New Rectangle(0, 0, intDiameter, intDiameter)

        ' Draw the circle.
        g.FillEllipse(brBee, rcBee)

        ' Free up resources.
        brBee.Dispose()

    End Sub

    Private Shared Sub DrawFeedingStation(ByVal g As Graphics, ByVal sngScaleFactor As Single, ByVal sngMagnification As Single, ByVal ptOrigin As PointF, ByVal ptThis As PointF)
        ' This accesses the brush we use to draw the feeding station.
        Dim brStation As Brush
        ' These store the dimensions of the feeder.
        Dim intEdge As Integer
        Dim rcStation As Rectangle

        ' Get the dimensions of the feeding station.
        ' N.B. We have to convert from millimetres to metres.
        intEdge = CType(Math.Ceiling((mcintFeedingStationEdge / 1000) * (sngMagnification / sngScaleFactor)), Integer)
        rcStation = New Rectangle(0, 0, intEdge, intEdge)

        ' Adjust the starting position, so that the feeder's location
        ' will be based on the centre rather than the bottom-left corner.
        ' So, we want the rectangle's bottom-left corner to be further over to the left and further down.
        ptThis.X -= CType(intEdge / 2, Single)
        ptThis.Y -= CType(intEdge / 2, Single)

        ' Adjust the co-ordinate transformation matrix.
        AdjustCTM(g, sngScaleFactor, ptOrigin, ptThis)

        ' Draw the square.
        brStation = New SolidBrush(Color.Green)
        g.FillRectangle(brStation, rcStation)
        brStation.Dispose()

    End Sub

    ' Calculate the magnification to use.
    ' The idea is for all the objects (bees/food/hives) to be displayed at the same scale,
    ' i.e. to be proportional to each other.
    ' However, this may not match the scale of the map, if they would be too small to be seen.
    ' E.g. if we have 1 pixel per metre, we might also use 1 pixel to represent a 2cm long bee.
    ' If we aren't displaying bees, we only have to get feeding stations and hives in synch, so we can make them smaller.
    Private Shared Function GetMagnification(ByVal sngScaleFactor As Single, ByVal blnShowBees As Boolean) As Single
        ' This stores the smallest length that we're dealing with.
        Dim intSmallestLength As Integer
        ' This stores the corresponding apparent length, i.e. the number of pixels required to display it.
        Dim sngSmallestPixels As Single
        ' This stores the magnification to use.
        Dim sngMagnification As Single

        ' Get the smallest length that we're dealing with, in millimetres.
        intSmallestLength = Math.Min(mcintHiveLength, mcintHiveWidth)
        intSmallestLength = Math.Min(intSmallestLength, mcintFeedingStationEdge)
        If blnShowBees Then
            intSmallestLength = Math.Min(intSmallestLength, mcintDroneLength)
            intSmallestLength = Math.Min(intSmallestLength, mcintQueenLength)
            intSmallestLength = Math.Min(intSmallestLength, mcintWorkerLength)
        End If

        ' Find out how many pixels this will take up.
        ' (Convert from millimetres to metres, then multiply by pixels per metre.)
        sngSmallestPixels = sngScaleFactor * intSmallestLength / 1000

        ' See how much we need to adjust it to make it visible.
        ' (If it's already visible, we don't shrink it, we just leave the magnification as 1.)
        If (sngSmallestPixels < 1) Then
            sngMagnification = 1 / sngSmallestPixels
        Else
            sngMagnification = 1
        End If

        ' Return the magnification as the function's result.
        Return sngMagnification

    End Function

    ' Scale factor - measured in pixels per metre.
    Public Shared Sub DisplayDome(ByVal g As Graphics, ByVal tsThis As CTimeSlice, ByVal blnShowBees As Boolean, ByVal sngScaleFactor As Single, ByVal ptOrigin As PointF)
        ' This is used to loop through the lists.
        Dim intLoop As Integer
        ' This accesses each bee in turn.
        Dim beeThis As SBee
        ' This stores the magnification to use.
        Dim sngMagnification As Single

        ' Set the appropriate background colour.
        ' g.Clear(Color.Black)

        ' Get the magnification.
        sngMagnification = GetMagnification(sngScaleFactor, blnShowBees)

        ' Draw the hives.
        For intLoop = 0 To tsThis.HiveCount - 1 Step 1
            DrawHive(g, sngScaleFactor, sngMagnification, ptOrigin, tsThis.Hives(intLoop))
        Next intLoop

        ' Draw the feeding stations.
        For intLoop = 0 To tsThis.FoodSourceCount - 1 Step 1
            DrawFeedingStation(g, sngScaleFactor, sngMagnification, ptOrigin, tsThis.FoodSources(intLoop))
        Next intLoop

        ' Draw the bees, if appropriate.
        If blnShowBees Then
            For intLoop = 0 To tsThis.BeeCount - 1 Step 1
                beeThis = tsThis.Bees(intLoop)
                DrawBee(g, sngScaleFactor, sngMagnification, ptOrigin, beeThis.Location, beeThis.Motion, beeThis.Caste)
                beeThis = Nothing
            Next intLoop
        End If

    End Sub
#End Region

#Region "Comb view"
    Private Shared Sub DrawQueen(ByVal g As Graphics)
        Dim brBlack As Brush = New SolidBrush(Color.Black)
        Dim brYellow As Brush = New SolidBrush(Color.Yellow)
        Dim penBlack As Pen = New Pen(Color.Black, 1)
        Dim penBlack2 As Pen = New Pen(Color.Black, 0.5)
        Dim penYellow As Pen = New Pen(Color.Yellow, 0.7)
        Dim rcThis As Rectangle

        ' Abdomen.
        rcThis = New Rectangle(0, 7, 13, 5)
        g.FillEllipse(brYellow, rcThis)

        ' Thorax.
        rcThis = New Rectangle(12, 7, 7, 5)
        g.FillEllipse(brBlack, rcThis)
        g.DrawEllipse(penYellow, rcThis)

        ' Eyes.
        rcThis = New Rectangle(18, 8, 2, 3)
        g.FillEllipse(brBlack, rcThis)

        ' Legs.
        g.DrawLine(penBlack, 3, 3, 12, 8)
        g.DrawLine(penBlack, 3, 17, 12, 12)
        g.DrawLine(penBlack, 15, 1, 14, 8)
        g.DrawLine(penBlack, 15, 19, 14, 12)
        g.DrawLine(penBlack, 22, 3, 18, 7)
        g.DrawLine(penBlack, 22, 16, 18, 12)

        ' Antennae.
        g.DrawLine(penBlack2, 22, 7, 20, 9)
        g.DrawLine(penBlack2, 22, 13, 20, 11)

#If cccBoundingCircle Then
        ' Bounding circle.
        rcThis = New Rectangle(0, 0, mcintQueenLength, mcintQueenLength)
        g.DrawEllipse(penBlack2, rcThis)
#End If

        ' Get rid of the graphic objects.
        brBlack.Dispose()
        brYellow.Dispose()
        penBlack.Dispose()
        penBlack2.Dispose()
        penYellow.Dispose()

    End Sub

    Private Shared Sub DrawDrone(ByVal g As Graphics)
        Dim brBlack As Brush = New SolidBrush(Color.Black)
        Dim brYellow As Brush = New SolidBrush(Color.Wheat)
        Dim penBlack As Pen = New Pen(Color.Black, 1)
        Dim penBlack2 As Pen = New Pen(Color.Black, 0.5)
        Dim rcThis As Rectangle

        ' Abdomen.
        rcThis = New Rectangle(0, 5, 9, 7)
        g.FillEllipse(brYellow, rcThis)

        ' Thorax.
        rcThis = New Rectangle(8, 5, 7, 7)
        g.FillEllipse(brYellow, rcThis)

        ' Eyes.
        rcThis = New Rectangle(15, 6, 2, 5)
        g.FillEllipse(brBlack, rcThis)

        ' Legs.
        g.DrawLine(penBlack, -2, 2, 7, 6)
        g.DrawLine(penBlack, -2, 15, 7, 11)
        g.DrawLine(penBlack, 9, 0, 10, 5)
        g.DrawLine(penBlack, 9, 17, 10, 12)
        g.DrawLine(penBlack, 19, 4, 14, 7)
        g.DrawLine(penBlack, 19, 13, 14, 10)

        ' Antennae.
        g.DrawLine(penBlack2, 20, 5, 17, 8)
        g.DrawLine(penBlack2, 20, 12, 17, 9)

#If cccBoundingCircle Then
        ' Bounding circle.
        rcThis = New Rectangle(0, 0, mcintDroneLength, mcintDroneLength)
        g.DrawEllipse(penBlack2, rcThis)
#End If

        ' Get rid of the graphic objects.
        brBlack.Dispose()
        brYellow.Dispose()
        penBlack.Dispose()
        penBlack2.Dispose()

    End Sub

    Private Shared Sub DrawWorker(ByVal g As Graphics)
        Dim brBlack As Brush = New SolidBrush(Color.Black)
        Dim brYellow As Brush = New SolidBrush(Color.LightYellow)
        Dim penBlack As Pen = New Pen(Color.Black, 1)
        Dim penBlack2 As Pen = New Pen(Color.Black, 0.5)
        Dim rcThis As Rectangle

        ' Abdomen.
        rcThis = New Rectangle(0, 5, 10, 5)
        g.FillEllipse(brYellow, rcThis)

        ' Thorax.
        rcThis = New Rectangle(9, 5, 5, 5)
        g.FillEllipse(brYellow, rcThis)

        ' Eyes.
        rcThis = New Rectangle(14, 6, 2, 3)
        g.FillEllipse(brBlack, rcThis)

        ' Legs.
        g.DrawLine(penBlack, 3, 2, 9, 6)
        g.DrawLine(penBlack, 3, 13, 9, 9)
        g.DrawLine(penBlack, 12, 2, 10, 6)
        g.DrawLine(penBlack, 12, 13, 10, 9)
        g.DrawLine(penBlack, 15, 3, 13, 5)
        g.DrawLine(penBlack, 15, 12, 13, 10)

        ' Antennae.
        g.DrawLine(penBlack2, 19, 5, 16, 7)
        g.DrawLine(penBlack2, 19, 10, 16, 8)

#If cccBoundingCircle Then
        ' Bounding circle.
        rcThis = New Rectangle(0, 0, mcintWorkerLength, mcintWorkerLength)
        g.DrawEllipse(penBlack2, rcThis)
#End If

        ' Get rid of the graphic objects.
        brBlack.Dispose()
        brYellow.Dispose()
        penBlack.Dispose()
        penBlack2.Dispose()

    End Sub

    ' See: http://www.gamedev.net/reference/articles/article1800.asp for the maths behind this.
    Private Shared Function GetHexagon(ByVal intSideLength As Integer) As Drawing2D.GraphicsPath
        ' These store the mathematical characteristics.
        Dim intA As Integer
        Dim intB As Integer
        Dim intH As Integer
        Dim intR As Integer
        ' This stores the number of points in the path.
        Const cintPointCount As Integer = 6
        ' This stores the list of points.
        Dim ptVertices(cintPointCount - 1) As Point
        ' This stores the types of the points.
        Dim pptTypes(cintPointCount - 1) As Byte    ' Drawing2D.PathPointType
        ' This is used to loop through the list of points.
        Dim intThisPoint As Integer
        ' This accesses the new graphics path.
        Dim gpThis As Drawing2D.GraphicsPath

        ' Calculate the mathematical characteristics.
        ' N.B. We need to convert from degrees to radians.
        intH = CType(Math.Sin(Math.PI / 6) * intSideLength, Integer)
        intR = CType(Math.Cos(Math.PI / 6) * intSideLength, Integer)
        intA = 2 * intR
        intB = intSideLength + (2 * intH)

        ' Get the co-ordinates.
        ptVertices(0) = New Point(intR, 0)                      ' top
        ptVertices(1) = New Point(intA, intH)                   ' top-right
        ptVertices(2) = New Point(intA, intSideLength + intH)   ' bottom-right
        ptVertices(3) = New Point(intR, intB)                   ' bottom
        ptVertices(4) = New Point(0, intSideLength + intH)      ' bottom-left
        ptVertices(5) = New Point(0, intH)                      ' top-left

        For intThisPoint = 0 To cintPointCount - 1 Step 1
            pptTypes(intThisPoint) = CType(Drawing2D.PathPointType.Line, Byte)
        Next intThisPoint
        pptTypes(0) = CType(pptTypes(0) Or Drawing2D.PathPointType.Start, Byte)
        pptTypes(cintPointCount - 1) = CType(pptTypes(cintPointCount - 1) Or Drawing2D.PathPointType.CloseSubpath, Byte)
        gpThis = New Drawing2D.GraphicsPath(ptVertices, pptTypes)
        Erase ptVertices
        Erase pptTypes

        ' Return a reference to the graphics path as the function's result.
        Return gpThis

    End Function

    Private Shared Sub DrawComb(ByVal g As Graphics, ByVal sngScaleFactor As Single, ByVal cmbThis As CComb)
        Dim intIndex As Integer = 0

        For intIndex = 0 To cmbThis.CellCount - 1 Step 1
            DrawCell(g, sngScaleFactor, cmbThis.Cells(intIndex))
        Next intIndex
    End Sub

    Private Shared Sub DrawCell(ByVal g As Graphics, ByVal sngScaleFactor As Single, ByVal celThis As CCell)
        Dim penBlack As Pen = New Pen(Color.Black, 0.5)
        Dim gpCell As Drawing2D.GraphicsPath
        Dim colFill As Drawing.Color
        Dim brFill As Drawing.SolidBrush

        ' Do the transformation.
        g.ResetTransform()
        g.ScaleTransform(sngScaleFactor, sngScaleFactor)
        g.TranslateTransform(celThis.Location.X, celThis.Location.Y)

        ' Set up the path (a hexagon).
        ' Specify sides of length 10mm.
        gpCell = GetHexagon(10)

        ' Draw the border of the cell.
        g.DrawPath(penBlack, gpCell)

        ' Draw the contents of the cell (if any).
        ' Nectar/pollen = solid block of appropriate colour.
        ' TODO: Larva = slightly more detailed shape. Different image based on age?
        Select Case celThis.Contents
            Case ECellContents.ccEgg, ECellContents.ccLarva, ECellContents.ccPupa
                colFill = Color.MediumVioletRed
            Case ECellContents.ccNectar
                colFill = Color.Aqua
            Case ECellContents.ccPollen
                colFill = Color.Lime
        End Select
        brFill = New SolidBrush(colFill)
        g.FillPath(brFill, gpCell)

        ' Get rid of the graphic objects.
        brFill.Dispose()
        penBlack.Dispose()
        gpCell.Dispose()

    End Sub

    Public Shared Sub DisplayComb(ByVal g As Graphics, ByVal tsThis As CTimeSlice, ByVal sngScaleFactor As Single)
        ' This is used to loop through the list of bees.
        Dim intThisBee As Integer
        ' This accesses each bee in turn.
        Dim beeThis As SBee

        ' Set the appropriate background colour.
        ' g.Clear(Color.White)

        ' Draw the comb.
        DrawComb(g, sngScaleFactor, tsThis.Comb)

        ' Draw the bees on this comb.
        For intThisBee = 0 To tsThis.BeeCount - 1 Step 1

            ' Access this bee.
            beeThis = tsThis.Bees(intThisBee)

            ' Do the transformation.
            g.ResetTransform()
            g.ScaleTransform(sngScaleFactor, sngScaleFactor)
            g.TranslateTransform(beeThis.Location.X, beeThis.Location.Y)
            g.RotateTransform(CType(beeThis.Direction * 180 / Math.PI, Single))

            ' Draw this bee.
            ' N.B. We don't use a "DrawBee" method, to avoid confusion with the dome-view version.
            Select Case beeThis.Caste
                Case ECaste.cstDrone
                    DrawDrone(g)
                Case ECaste.cstQueen
                    DrawQueen(g)
                Case ECaste.cstWorker
                    DrawWorker(g)
            End Select

            ' Free up the object reference.
            beeThis = Nothing

            ' Move to the next bee.
        Next intThisBee

    End Sub
#End Region

End Class
