Option Compare Binary
Option Explicit On 
Option Strict On

Public Class CDome

    ' The current date and time in the simulation (distinct from the physical computer's date/time).
    Private Shared mdtmCurrent As Date

    ' The simulated season (based on the date).
    ' N.B. We are basically storing the same information twice here, since the season can be derived from the date.
    ' This is for efficiency, to cache the result and avoid the overhead of performing the same calculation several times per tick.
    Private Shared mssnSeason As ESeason

    ' The hives.
    Private Shared mintHiveCount As Integer = 0
    Private Shared mhvHives As CHive()

    ' The food sources.
    Private Shared mintFoodSourceCount As Integer
    Private Shared mfsFoodSources As CFoodSource()

    ' The outdoor bees.
    Private Shared mintBeeCount As Integer
    Private Shared mbeeBees As CBee()

    ' This stores the quantities, for speed of access (rather than recounting all the bees each time we need this info).
    Private Shared mqntQuantities As SQuantities

    ' The dome temperature (in degrees Kelvin).
    Private Shared mintTemperature As Integer = 288

    ' The interval between ticks, in real time and simulation time.
    ' These are both measured in milliseconds.
    Private Shared mdblRealTickInterval As Double = 0
    Private Shared mdblSimTickInterval As Double = 0

    ' This stores the total elapsed simulation time (in milliseconds).
    Private Shared mdblSimElapsedTime As Double = 0

    ' Use these variables to work out other values, e.g. speed and life expectancy.
    ' Ideally they would be kept proportional to each other, i.e. if one gets doubled then so does the other.
    ' That way, their effect would be cancelled out for calculations involving distance etc.
    ' You would still get accelerated time, but it would really be the equivalent of watching something on fast-forward,
    ' whereas if you increase the time-warp factor without increasing the ticks per second then you get
    ' more of an approximation, e.g. bees that will fly further in each tick, distorting their search pattern
    ' (fewer changes in direction over the same distance).
    ' The downside is that (as noted below), increasing the rate of ticks is computationally expensive,
    ' so the speed of the hardware will be a limiting factor on this.
    '
    ' N.B. There is also a problem with the granularity of the DateTime class.
    ' In theory (according to the documentation) it goes down to 100 nanoseconds, i.e. 0.1 milliseconds.
    ' In practice, I have found that when I have 10,000 ticks per second (i.e. 1 tick every 0.1 milliseconds),
    ' I only actually get about 1100 ticks in a 15 second period, rather than 150,000.
    ' Let's say that there's a realistic limit of 1000 ticks per second.

    ' One tick every 100ms = 10 ticks per second.
    ' A higher value means smoother motion, but also requires a faster computer.
    Private Shared mintTicksPerSecond As Integer

    ' Are we running in real-time, or faster?
    ' N.B. In theory we could run slower, but there's no real benefit to that;
    ' the purpose of running faster is to avoid waiting for a year to get a year's worth of results from the model.
    ' 1 = real-time
    ' 2 = twice real-time (2 seconds in simulation = 1 second in observer's time-frame)
    ' 0.5 = half real-time (1 second in simulation = 2 seconds in observer's time-frame)
    ' etc.
    ' To get a year in one day, you would need a factor of about 365.
    ' To get a year in one hour, about 365 * 24 = 8760.
    Private Shared msngTimeWarpFactor As Single

    ' The speed multiplier, based on distance per tick.
    ' This is normally TimeWarpFactor/TicksPerSecond, but initially it is 1 (during calibration),
    ' since those other values haven't been set yet.
    ' Ideally (with an infinitely fast computer), this would be 1.
    ' N.B. This isn't acceleration - the speed gets adjusted initially, then stays the same over time.
    Private Shared mdblSpeedMultiplier As Double = 1

    ' Real-time: 7 metres per second, 10 ticks per second, therefore 0.7 metres per tick.
    ' More ticks: 7 metres per second, 100 ticks per second, therefore 0.07 metres per tick.
    ' Double-time: 14 metres per second, 10 ticks per second, therefore 1.4 metres per tick.
    ' Double-time, double ticks: 14 metres per second, 20 ticks per second, therefore 0.7 metres per tick.

    ' This stores the number of ticks per day.
    Private Shared mintTicksPerDay As Integer = 0

    ' TODO: Support different starting scenarios, including a fairly complex one for calibration (to test limits).
    Public Shared Sub Initialise()
        ' This is used to loop through the list of hives.
        Dim intThisHive As Integer
        ' This stores each hive's quantities.
        Dim qntHive As SQuantities
        ' This is used to loop through the list of bees.
        Dim intThisBee As Integer

        ' Store the current date and time.
        mdtmCurrent = Now()

        ' Set all the starting quantities to zero.
        With mqntQuantities
            .DroneCount = 0
            .BroodDroneCount = 0
            .BroodQueenCount = 0
            .BroodWorkerCount = 0
            .QueenCount = 0
            .WorkerCount = 0
        End With

        ' Put a hive in the middle of the dome.
        mintHiveCount = 1
        ReDim mhvHives(0)
        mhvHives(0) = New CHive(New PointF(0, 0), 1)

        ' Get each hive's quantities, and add them to the dome's total,
        ' then add an event handler for changes to these quantities.
        For intThisHive = 0 To mintHiveCount - 1 Step 1
            mhvHives(intThisHive).GetQuantities(qntHive)
            With mqntQuantities
                .DroneCount += qntHive.DroneCount
                .BroodDroneCount += qntHive.BroodDroneCount
                .BroodQueenCount += qntHive.BroodQueenCount
                .BroodWorkerCount += qntHive.BroodWorkerCount
                .QueenCount += qntHive.QueenCount
                .WorkerCount += qntHive.WorkerCount
            End With
            AddEventHandlers(mhvHives(intThisHive))
        Next intThisHive

        ' Outdoor bees.
        mintBeeCount = 40
        ReDim mbeeBees(mintBeeCount - 1)
        For intThisBee = 0 To mintBeeCount - 1 Step 1
            mbeeBees(intThisBee) = New CWorker(mhvHives(0).Location, New PointF(0, 0), 1, True)
            AddEventHandlers(mbeeBees(intThisBee))
        Next intThisBee
        mqntQuantities.WorkerCount += mintBeeCount

        ' Food sources (feeding stations).
        mintFoodSourceCount = 4
        ReDim mfsFoodSources(mintFoodSourceCount - 1)
        ' Co-ordinates given in metres from hive.
        mfsFoodSources(0) = New CFeedingStation(New PointF(-200, -200), 10)
        mfsFoodSources(1) = New CFeedingStation(New PointF(-200, 200), 20)
        mfsFoodSources(2) = New CFeedingStation(New PointF(200, -200), 30)
        mfsFoodSources(3) = New CFeedingStation(New PointF(200, 200), 40)

    End Sub

    Public Shared Property CurrentDateTime() As Date
        Get
            Return mdtmCurrent
        End Get
        Set(ByVal Value As Date)
            mdtmCurrent = Value
        End Set
    End Property

    Public Shared ReadOnly Property BeeCount() As Integer
        Get
            Return mintBeeCount
        End Get
    End Property

    Public Shared ReadOnly Property Bees(ByVal intIndex As Integer) As CBee
        Get
            Return mbeeBees(intIndex)
        End Get
    End Property

    Public Shared ReadOnly Property FoodSourceCount() As Integer
        Get
            Return mintFoodSourceCount
        End Get
    End Property

    Public Shared ReadOnly Property FoodSources(ByVal intIndex As Integer) As CFoodSource
        Get
            Return mfsFoodSources(intIndex)
        End Get
    End Property

    Public Shared Function GetHive(ByVal intHiveID As Integer) As CHive
        ' This is used to loop through the list of hives.
        Dim intThisHive As Integer
        ' This accesses the relevant hive.
        Dim hvReturnValue As CHive

        ' Look for this hive, and stop if we find it.
        intThisHive = 0
        Do While (intThisHive < mintHiveCount) And (hvReturnValue Is Nothing)
            If (mhvHives(intThisHive).ID = intHiveID) Then
                hvReturnValue = mhvHives(intThisHive)
            Else
                intThisHive += 1
            End If
        Loop

        ' Return a reference to the hive (if we found it) as the function's result.
        Return hvReturnValue

    End Function

    Public Shared ReadOnly Property HiveCount() As Integer
        Get
            Return mintHiveCount
        End Get
    End Property

    Public Shared ReadOnly Property Hives(ByVal intIndex As Integer) As CHive
        Get
            Return mhvHives(intIndex)
        End Get
    End Property

    Public Shared ReadOnly Property Season() As ESeason
        Get
            Return mssnSeason
        End Get
    End Property

    ' The temperature in degrees Kelvin.
    Public Shared Property TemperatureKelvin() As Integer
        Get
            Return mintTemperature
        End Get
        Set(ByVal Value As Integer)
            mintTemperature = Value
        End Set
    End Property

    ' The temperature in degrees Celsius.
    Public Shared ReadOnly Property TemperatureCelsius() As Integer
        Get
            Return mintTemperature - 273
        End Get
    End Property

    ' Assign a time-slice's properties in dome view.
    Public Overloads Shared Sub AssignTimeSlice(ByVal tsThis As CTimeSlice)
        ' This is used to loop through the lists.
        Dim intLoop As Integer
        ' This accesses each bee in turn.
        Dim beeThis As SBee

        ' Assign the common properties.
        AssignTimeSliceCommon(tsThis)

        ' Assign the dome-specific properties.

        ' Dome view.
        tsThis.DomeView = True

        ' Hives.
        tsThis.HiveCount = mintHiveCount
        For intLoop = 0 To mintHiveCount - 1 Step 1
            tsThis.Hives(intLoop) = mhvHives(intLoop).Location
        Next intLoop

        ' Food sources.
        tsThis.FoodSourceCount = mintFoodSourceCount
        For intLoop = 0 To mintFoodSourceCount - 1 Step 1
            tsThis.FoodSources(intLoop) = mfsFoodSources(intLoop).Location
        Next intLoop

        ' Bees.
        ' N.B. The time-slice contains structures rather than objects,
        ' so we are copying around the contents rather than the reference.
        tsThis.BeeCount = mintBeeCount
        For intLoop = 0 To mintBeeCount - 1 Step 1
            With mbeeBees(intLoop)
                beeThis.Caste = .Caste
                beeThis.Direction = .Direction
                beeThis.Location = .Location
                beeThis.Motion = .Motion
            End With
            tsThis.Bees(intLoop) = beeThis
        Next intLoop

    End Sub

    ' Assign a time-slice's properties in hive view.
    Public Overloads Shared Sub AssignTimeSlice(ByVal tsThis As CTimeSlice, ByVal intHiveIndex As Integer, ByVal intFrameIndex As Integer)

        ' Assign the common properties.
        AssignTimeSliceCommon(tsThis)

        ' Assign the hive-specific properties.
        mhvHives(intHiveIndex).AssignTimeSlice(tsThis, intFrameIndex)

    End Sub

    ' This assigns the time slice properties that are common to both views.
    Private Shared Sub AssignTimeSliceCommon(ByVal tsThis As CTimeSlice)

        ' Simulated date/time.
        tsThis.SimulatedDateTime = mdtmCurrent

        ' Dome temperature.
        tsThis.TemperatureKelvin = mintTemperature

        ' Quantities.
        tsThis.Quantities = mqntQuantities

    End Sub

    Public Shared Sub Tick()
        ' This is used to loop through the list of hives.
        Dim intThisHive As Integer
        ' This is used to loop through the list of bees.
        Dim intThisBee As Integer

        ' Increment the shared date and time.
        ' N.B. This is different to the computer's date and time.
        mdtmCurrent = mdtmCurrent.AddMilliseconds(mdblSimTickInterval)
        mdblSimElapsedTime += mdblSimTickInterval

        ' Calculate the season from this date.
        Select Case Month(mdtmCurrent)
            Case 2 To 4
                mssnSeason = ESeason.ssnSpring
            Case 5 To 7
                mssnSeason = ESeason.ssnSummer
            Case 8 To 9
                mssnSeason = ESeason.ssnAutumn
            Case 10 To 12, 1
                mssnSeason = ESeason.ssnWinter
        End Select

        ' Assign an appropriate dome temperature.
        AssignTemperature()

        ' Do the hives' ticks (i.e. handle any bees that are inside).
        For intThisHive = 0 To mintHiveCount - 1 Step 1
            mhvHives(intThisHive).Tick()
        Next intThisHive

        ' Let the outdoor bees tick.
        ' N.B. The size of this array may change during the execution of the loop (e.g. if bees die),
        ' so we can't do a conventional For loop in case we wind up out of bounds.
        ' There is a risk that some bees will be overlooked, or tick twice, but I can live with that.
        intThisBee = 0
        Do While (intThisBee < mintBeeCount)
            mbeeBees(intThisBee).Tick()
            intThisBee += 1
        Loop

    End Sub

    Private Shared Sub AssignTemperature()
        ' This stores the monthly average temperature.
        Dim sngMonthlyAverage As Single

        ' Assign the current dome temperature, based on the simulated date/time.

        ' See what month this is, and get the average monthly temperature in degrees Celsius (based on London).
        ' Source of information: http://www.worldclimate.com/cgi-bin/data.pl?ref=N51W000+1202+0005305G2
        Select Case Month(mdtmCurrent)
            Case 1
                sngMonthlyAverage = 3.8
            Case 2
                sngMonthlyAverage = 3.9
            Case 3
                sngMonthlyAverage = 5.9
            Case 4
                sngMonthlyAverage = 7.9
            Case 5
                sngMonthlyAverage = 11.2
            Case 6
                sngMonthlyAverage = 14.3
            Case 7
                sngMonthlyAverage = 16.5
            Case 8
                sngMonthlyAverage = 16.2
            Case 9
                sngMonthlyAverage = 13.8
            Case 10
                sngMonthlyAverage = 10.8
            Case 11
                sngMonthlyAverage = 6.6
            Case 12
                sngMonthlyAverage = 4.7
        End Select

        ' Adjust it based on the time of day.
        ' Split the day into three 8-hour periods:
        ' 08:00->16:00 (the 8 hours surrounding noon) is the hottest - 3 degrees above average.
        ' 04:00->08:00, 16:00->20:00 (the 8 hours surrounding that) is average.
        ' 00:00->04:00, 20:00->00:00 (the 8 hours surrounding midnight) is the coldest - 3 degrees below average.
        Select Case Hour(mdtmCurrent)
            Case 8 To 16
                sngMonthlyAverage += 3
            Case 0 To 4, 20 To 23
                sngMonthlyAverage -= 3
        End Select

        ' Now convert it to Kelvins.
        sngMonthlyAverage += 273

        ' Store this as the current dome temperature.
        mintTemperature = CType(sngMonthlyAverage, Integer)

    End Sub

    Public Shared Property RealTickInterval() As Double
        Get
            Return mdblRealTickInterval
        End Get
        Set(ByVal Value As Double)
            mdblRealTickInterval = Value
        End Set
    End Property

    Public Shared Property SimTickInterval() As Double
        Get
            Return mdblSimTickInterval
        End Get
        Set(ByVal Value As Double)
            mdblSimTickInterval = Value
        End Set
    End Property

    Public Shared Property SimElapsedTime() As Double
        Get
            Return mdblSimElapsedTime
        End Get
        Set(ByVal Value As Double)
            mdblSimElapsedTime = Value
        End Set
    End Property

    Public Shared Property SpeedMultiplier() As Double
        Get
            Return mdblSpeedMultiplier
        End Get
        Set(ByVal Value As Double)
            mdblSpeedMultiplier = Value
        End Set
    End Property

    Public Shared ReadOnly Property TicksPerDay() As Integer
        Get
            Return mintTicksPerDay
        End Get
    End Property

    Public Shared Property TicksPerSecond() As Integer
        Get
            Return mintTicksPerSecond
        End Get
        Set(ByVal Value As Integer)
            mintTicksPerSecond = Value
            mintTicksPerDay = mintTicksPerSecond * 60 * 60 * 24
        End Set
    End Property

    Public Shared Property TimeWarpFactor() As Single
        Get
            Return msngTimeWarpFactor
        End Get
        Set(ByVal Value As Single)
            msngTimeWarpFactor = Value
        End Set
    End Property

    Public Shared Sub GetQuantities(ByRef qntQuantities As SQuantities)

        ' Return the quantities.
        ' N.B. Since this is a structure rather than a class,
        ' we can just copy the values straight across.
        qntQuantities = mqntQuantities

    End Sub

    Public Shared Sub QuantityChanged(ByVal cstCaste As ECaste, ByVal blnBrood As Boolean, ByVal intDifference As Integer)

        ' Update the relevant quantity.
        CShared.QuantityChanged(cstCaste, blnBrood, intDifference, mqntQuantities)

    End Sub

    Private Shared Sub DeadBee(ByVal beeThis As CBee)
        ' This accesses the bee's home hive.
        Dim hvHome As CHive

        ' This bee should simply disappear (no other agent based code required).

        ' Reduce the total number of bees.
        CShared.QuantityChanged(beeThis.Caste, False, -1, mqntQuantities)

        ' Also reduce the quantity in this bee's home hive.
        hvHome = GetHive(beeThis.HiveID)
        hvHome.LostBee(beeThis.Caste, False, -1)
        hvHome = Nothing

        ' Remove this bee from the array.
        CShared.RemoveBee(beeThis, mintBeeCount, mbeeBees)

        ' Remove this bee's event handlers.
        RemoveEventHandlers(beeThis)

    End Sub

    ' This handles events when bees leave the general dome area (to enter a hive).
    Private Shared Sub BeeLeftDome(ByVal beeThis As CBee)

        ' Remove this bee from the array.
        CShared.RemoveBee(beeThis, mintBeeCount, mbeeBees)

        ' Technically, the total number of bees in the simulation hasn't changed.
        ' However, when the bee enters a hive, the quantity in that hive will go up.
        ' This will affect the global quantities, so we need to decrement them here to compensate.
        CShared.QuantityChanged(beeThis.Caste, False, -1, mqntQuantities)

        ' Remove this bee's event handlers.
        RemoveEventHandlers(beeThis)

    End Sub

    ' This is used when a bee flies outside a hive, and comes under the dome's jurisdiction.
    Public Shared Sub AddBee(ByVal beeThis As CBee)

        ' Make room for this bee in the array.
        ReDim Preserve mbeeBees(mintBeeCount)

        ' Store a reference to this bee in the array.
        mbeeBees(mintBeeCount) = beeThis

        ' Increment the number of bees.
        mintBeeCount += 1

        ' Technically, the total number of bees in the simulation hasn't changed.
        ' However, when the bee leaves its hive, the quantity in that hive will go down.
        ' This will affect the global quantities, so we need to increment them here to compensate.
        CShared.QuantityChanged(beeThis.Caste, False, 1, mqntQuantities)

        ' Add the event handlers for this bee.
        AddEventHandlers(beeThis)

    End Sub

    ' Add the event handlers for this hive.
    Private Overloads Shared Sub AddEventHandlers(ByVal hvThis As CHive)

        AddHandler hvThis.QuantityChanged, AddressOf QuantityChanged

    End Sub

    ' Add the event handlers for this bee.
    Private Overloads Shared Sub AddEventHandlers(ByVal beeThis As CBee)

        AddHandler beeThis.DeadBee, AddressOf DeadBee
        AddHandler beeThis.LeftDome, AddressOf BeeLeftDome

    End Sub

    ' Remove the event handlers for this bee.
    Private Shared Sub RemoveEventHandlers(ByVal beeThis As CBee)

        RemoveHandler beeThis.DeadBee, AddressOf DeadBee
        RemoveHandler beeThis.LeftDome, AddressOf BeeLeftDome

    End Sub

    ' If a hive's queen is in the dome area, kill it, and return True.
    ' If the queen isn't here, return False.
    ' N.B. This is very similar to the function in CComb, but it takes a hive ID as a parameter.
    Public Shared Function KillQueen(ByVal intHiveID As Integer) As Boolean
        ' This stores the function's return value.
        Dim blnReturnValue As Boolean = False
        ' This is used to loop through the list of bees.
        Dim intThisBee As Integer
        ' This accesses the bee to delete.
        Dim beeThis As CBee

        ' Loop through the list of bees looking for the queen.
        ' (Stop if we find it, for efficiency.)
        intThisBee = 0
        Do While (intThisBee < mintBeeCount) And Not blnReturnValue
            If ((mbeeBees(intThisBee).Caste = ECaste.cstQueen) And (mbeeBees(intThisBee).HiveID = intHiveID)) Then
                blnReturnValue = True
            Else
                intThisBee += 1
            End If
        Loop

        ' If we found it, delete it, and handle the clean-up afterwards.
        If blnReturnValue Then
            beeThis = mbeeBees(intThisBee)
            CShared.RemoveBee(intThisBee, mintBeeCount, mbeeBees)
            DeadBee(beeThis)
            beeThis = Nothing
        End If

        ' Return the appropriate result from this function.
        Return blnReturnValue

    End Function

End Class
